
/*
 *  regen - regenerate CVS source when predecessor sources are updated
 *
 *  $Id: regen.c,v 1.14 2000/09/11 02:26:04 bruce Exp $
 *
 *  The approach here is to compute the list of any derived files based on
 *  the provided list of files being committed.  If there are no such
 *  derived files, this program exits successfully.  All files being
 *  committed must be either all top of tree or all from the same
 *  branch tag.
 *
 *  If there are derived files, then the predecessor files are copied
 *  into a staging area.  The files being committed are copied and any
 *  other files are gotten via "cvs get" for the appropriate branch.
 *  This avoids the recursion lock problem.
 *
 *  A simplifying assumption we make is that if there are more than one
 *  predecessor files, then they all reside in the same directory and are
 *  thus either listed or not being committed at this time.  At the current
 *  time, there is only one example of a derived file that has multiple
 *  predecessor files:  gcc/fixinc/fixincl.x.  Both if its predecessor
 *  files (inclhack.def and fixincl.tpl) live in the same directory.
 *  (On a multi-directory commit, CVS will invoke this program once for
 *  each directory containing files to commit.  There is no simple method
 *  for determining the complete list of files being committed at once.
 *  A serious CVS deficiency.)
 */

#define DEFINING
#include "regen.h"
#include <pwd.h>

#ifdef DEBUG
# define DEFAULT_MAILTO  "bkorb@gnu.org"
#else
# define DEFAULT_MAILTO  "gcc-bugs@gcc.gnu.org"
#endif

#ifndef NOPROCESS
#define NOPROCESS       ((pid_t) -1)
#define NULLPROCESS     ((pid_t)0)
#endif

tSCC zFsErr[]      = "from line %d - regen FS err %d (%s):\n\t%s on %s\n";
tSCC zCmdFailFmt[] = "\nCOMMAND FAILURE:\n%s";
tSCC z_dash_A[]    = "-A";

#define MAX_FILES_IN_FMT   32

FILE*  log_fp = NULL;
char*  pz_b1;
char*  pz_b2;

/*
 *  Define a dependency descriptor structure we will use to
 *  wire all this stuff together
 */
typedef struct dep_desc t_dep_desc;
struct dep_desc {
    ulong_t  tag;         /* tag for dependent/target name    */
    tCC*     pz_name;
    tCC**    papz_assoc;  /* ptr to array of associated names */
    int      referenced;
};

static int map_ct = 0;
static int map_allocs = 0;

typedef struct map_info t_map_info;
struct map_info {
    size_t      map_size;
    tCC*        map_dta;
    int         map_fd;
};

static t_map_info*  p_map_info = NULL;

#include "regen.x"

void        cleanup(          void );
size_t      export_file_name( char* pz_buf, tCC* pz_file );
void        fail_exit(        tCC* pzMsg, int line );
t_dep_desc* find_dep_entry(   t_dep_desc* pTable, int entry_ct, tCC* pz_name );
t_dep_desc* find_predecessor( tCC* pz_file, tCC* pz_ver, size_t ver_len );
void        find_version(     const char* pz_file, const char** ppz_ver,
                              size_t* p_ver );
void        get_predecessor_files( tCC** papz_assoc );
t_bool      has_derivative(   tCC* pz_file );
void        initialize(       size_t max_file_name );
tCC*        map_data(         tCC* pz_file );
void        mark_targets(     tCC* pz_file, t_dep_desc* pDT );
void        rebuild_referenced_targets( void );
void        run_system(       tCC* pz_cmd, int line );
void        setup_work_dir(   tCC* pz_ver );


#define NO_MAIN
#include "hashstr.c"


    int
main( int     argc,
      char**  argv )
{
    if (--argc < 2) {
        tSCC zArgs[] =
            "regen ERROR:  need more than %d argument(s).\n"
            "USAGE:  regen <cvs-directory> <file-name> [ ... ]\n";

        fprintf( stderr, zArgs, argc );
        return EXIT_FAILURE;
    }

    /*
     *  Figure out the names of all the directories we are interested in
     *  and allocate buffers big enough to do all our scribbling.
     */
    pz_cvsroot = *(++argv);

    {
        size_t  maxlen = 0;
        char** av = ++argv;
        int    ct = --argc; /* we skip the cvs project directory */

        while (--ct >= 0) {
            char*  a = *av++;
            size_t l = strlen( a );
            if (l > maxlen)
                maxlen = l;
        }

        initialize( maxlen );
    }

    {
        int derived_count = 0;

        while (--argc >= 0)
            if (has_derivative( *(argv++) ))
                derived_count++;

        if (derived_count == 0) {
            fputs( "regen has nothing to do\n", stdout );
            return EXIT_SUCCESS;
        }
    }

    fflush( stdout );
    fflush( stderr );

#ifdef DEBUG
    rebuild_referenced_targets();
#else
    switch (fork()) {
    case NOPROCESS:
    fork_error:
        sprintf( pz_b2, "fs error %d (%s) on fork()\n",
                 errno, strerror( errno ));
        fail_exit( pz_b2, __LINE__ );
        /* NOTREACHED */

    default:
        pz_failure = NULL;
        break;

    case NULLPROCESS:
        /*
         *  Redirect all I/O to the log file and become a daemon
         */
        freopen( "/dev/null", "r", stdin );
        dup2( fileno( log_fp ), STDOUT_FILENO );
        dup2( fileno( log_fp ), STDERR_FILENO );
        setsid();

        switch (fork()) {
        case NOPROCESS:
            goto fork_error;

        case NULLPROCESS:
	    rebuild_referenced_targets();

        /* default:  parent falls through and returns SUCCESS */
        }
    }
#endif

    return EXIT_SUCCESS;
} /* END-MAIN - start alphabatized support routines */


/*
 *  cleanup
 *
 *  Check for failure conditions.  If none, then purge the work directory.
 *  If so, then emit email to let people know.
 */
    void
cleanup( void )
{
    tSCC zMail[] =
        "sendmail %1$s <<_EOF_\n"
        "To: CVS commit problems <%1$s>\n"
        "From: automated commission <%2$s>\n"
        "Subject: CVS Commit Oops in %3$s by %4$s\n\n"
        "**** Error message: ****\n\n"
        "`cat %6$s`\n"
        "**** Commit log: ****\n\n"
        "`cat %5$s/LOG`\n\n"
        "_EOF_\n";

#ifdef DEBUG
    tSCC zCleanupCmd[] = "echo rm -rf %s >&2";
#else
    tSCC zCleanupCmd[] = "rm -rf %s";
#endif

    if (log_fp != NULL)
        fclose( log_fp );

    /*
     *  IF pz_failure is not NULL,
     *  THEN we are the child process that did the work.
     */
    if (pz_failure != NULL) {
        if (access( pz_failure, R_OK ) == 0) {
            struct passwd * p = getpwuid( getuid() );

            char* pzErr = getenv( "CVS_ERRORS_TO" );

            if (pzErr == (char*)NULL)
                pzErr = DEFAULT_MAILTO;

            sprintf( pz_b1, zMail, pzErr, p->pw_name, pz_cvsroot,
                     p->pw_gecos, pz_work_dir, pz_failure );
            freopen( "/dev/null", "w", stdout );
            freopen( "/dev/null", "w", stderr );

        } else {
            sprintf( pz_b1, zCleanupCmd, pz_work_dir );
        }

        system( pz_b1 );
    }
}


/*
 *  export_file_name
 *
 *  Construct shell commands to assign a well-known variable name
 *  to the full path of one of the predecessor files.  The build rule
 *  scripts are allowed to use this name to find their source files.
 */
    size_t
export_file_name( char* pz_buf, tCC* pz_file )
{
    tSCC    zPfx[] = "FILE_";

    char*   pzOut;
    tCC*    pz = strrchr( pz_file, '/' );
    size_t  res;

    if (pz++ == NULL)
        pz = pz_file;

    strcpy( pz_buf, zPfx );
    pzOut = pz_buf + sizeof( zPfx ) - 1;

    while (*pz) {
        if (isalnum( *pz ))
            *(pzOut++) = toupper( *(pz++));
        else
            *(pzOut++) = (*(pz++), '_');
    }
    res = (pzOut - pz_buf);
    res += sprintf( pzOut, "='%s/%s'\n", pz_work_dir, pz_file );
    return res;
}


/*
 *  fail_exit
 *
 *  Put out lots of noise in the event of failure
 */
    void
fail_exit( tCC* pzMsg, int line )
{
    FILE* fp;
    if (pz_failure == NULL)
        pz_failure = "/tmp/FAILURE";

    fp = fopen( pz_failure, "a" );
    fprintf( fp, "\nregen ERROR from line %d:\n\t%s\n", line, pzMsg );
    fprintf( stderr, "regen ERROR from line %d:\n\t%s\n", line, pzMsg );
    fclose( fp );
    fclose( stderr );
    exit( EXIT_FAILURE );
    /* NOTREACHED */
}


/*
 *  find_dep_entry
 *
 *  Search a dependency table for an entry.  We assume that all hash
 *  results are unique.  If not, then modify "hashstr.c" and make it so.
 *  I should probably use "gperf" to ensure proper results, but a 32 bit
 *  hash on a couple dozen file names really ought to always be unique.
 */
    t_dep_desc*
find_dep_entry( t_dep_desc* pTable, int entry_ct, tCC* pz_name )
{
    /*
     *  Now, hunt for a well-formed string.
     *  IF we don't find it, then the file being committed is
     *  not a predecessor to any generated file.
     */
    int hi = entry_ct - 1;
    int lo = 0;
    ulong_t tag = hashstr( pz_name );

    for (;;) {
        int av = (hi + lo) / 2;
        t_dep_desc* pDD = pTable + av;

        if (pDD->tag == tag) {
            if (strcmp( pz_name, pDD->pz_name ) == 0)
                return pDD;
            return NULL;
        }

        if (pDD->tag < tag)
            lo = av + 1;
        else
            hi = av - 1;
        if (lo > hi)
            break;
    }
    return NULL;
}


/*
 *  is_predecessor
 *
 *   - see if a file is used as source to a generated file.
 */
    t_dep_desc*
find_predecessor( tCC* pz_file, tCC* pz_ver, size_t ver_len )
{
    tSCC*          pz_last_ver;
    static size_t  last_len;

    if (pz_ver_tag == (char*)NULL) {
        /*
         *  IF the user is using a symbolic tag,
         *  THEN make sure we use the same one
         */
        if (ver_len > 0) {
            char* pz = (char*)xmalloc( ver_len + 6 );
            pz_ver_tag = (const char*)pz;
            *(pz++) = '-';
            *(pz++) = 'r';
            *(pz++) = '\'';
            memcpy( (void*)pz, (void*)pz_ver, ver_len );
            pz += ver_len;
            *(pz++) = '\'';
            *(pz++) = '\0';
        }
        else
            pz_ver_tag = z_dash_A;

        pz_last_ver = pz_ver;
        last_len    = ver_len;
    }
    else if (  (ver_len != last_len)
            || (  (ver_len > 0)
               && (strncmp( pz_ver, pz_last_ver, ver_len ) != 0 )
            )  )  {
        sprintf( pz_b2, "inconsistent tagging\n" );
        fail_exit( pz_b2, __LINE__ );
    }

    return find_dep_entry( dep_to_targets, DEP_TO_TARGETS_CT, pz_file );
}


/*
 *  find_version
 *
 *  FOR every file checked in, ensure the version tags are either top of tree
 *  or all the same.  We save the first one in a global string.
 */
    void
find_version( const char* pz_file, const char** ppz_ver, size_t* p_ver )
{
    char z[ 1024 ];
    size_t skip_ct = sprintf( z, "/%s/", pz_file );
    const char* pz_find = (const char*)strstr( pz_entries, z );

    if (pz_find == NULL)
    {
        sprintf( pz_b2, "``%s'' is not in the Entries file\n",
                 pz_file );
        fail_exit( pz_b2, __LINE__ );
    }

    /*
     *  Skip over the next three fields + the terminating slashes
     */
    pz_find += skip_ct;
    skip_ct = 3;
    for (;;)
    {
        for (;;)
        {
            static const char zNoTag[] =
                "no tag entry for ``%s''\n%s\n";

            char ch = *(pz_find++);

            switch (ch)
            {
            case '\0':
            case '\n':
                sprintf( pz_b2, zNoTag, pz_file, pz_entries );
                fail_exit( pz_b2, __LINE__ );

            case '/':
                goto found_slash;
            }
        } found_slash:;

        if (--skip_ct == 0)
            break;
    }

    /*
     *  IF we do not have a symbolic tag,
     *  THEN we must be using top of tree
     */
    if ((pz_find[0] != 'T') || isdigit( pz_find[1] )) {
        *p_ver = 0;
        *ppz_ver = "";
        return;
    }

    /*
     *  Return the pointer to the version string and its length
     */
    *ppz_ver = ++pz_find;
    skip_ct = 0;
    while ((*pz_find != '\0') && (! isspace( *(pz_find++) )))
        skip_ct++;
    *p_ver = skip_ct;
}


/*
 *  get_predecessor_files
 *
 *  For each predecessor file of the current generated file,
 *  first, make sure the directory exists for it in our work area.
 *  Then if the file does *not* already exist (i.e. has been copied
 *  in from the original checkin directory), then "cvs get" it.
 */
    void
get_predecessor_files( tCC** papz_assoc )
{
    tCC*   pz_v = (pz_ver_tag == z_dash_A) ? "" : pz_ver_tag;
    tCC**  papz;
    char*  pzGetList = pz_b2;
    size_t off, len;

    FILE*  pin;
    FILE*  pout;

    /*
     *  For each predecessor file of the current generated file,
     *  make sure it either exists or extract it from the annotation
     */
    papz = papz_assoc;
    for (;;) {
        tCC*  pz_assoc = *(papz++);
        char* pz_trim;

        if (pz_assoc == NULL)
            break;

        /*
         *  Check for file being committed (already copied in)
         */
        if (access( pz_assoc, R_OK ) == 0)
            continue;

        /*
         *  add the file to the get list
         */
        pzGetList += sprintf( pzGetList, "%s ", pz_assoc );
    }

    /*
     *  Is this really necessary?  We may already have all we need.
     */
    if (pzGetList != pz_b2) {
        tSCC zGet[] =
#ifdef DEBUG
            "set -x\n"
#endif
            "cvs -d %s get %s %s || \\\n"
            "  echo get of predecessors FAILED > %s";

        sprintf( pz_b1, zGet, pz_cvsroot, pz_v, pz_b2, pz_failure );

        /*
         *  Ensure there was no failure exit of the annotate command
         */
        run_system( pz_b1, __LINE__ );
    }
}


/*
 *  has_derivative
 *
 *  Return B_TRUE if the file name passed in is a predecessor to a
 *  generated file.
 */
    t_bool
has_derivative( tCC* pz_file )
{
    static t_bool setup_needed = B_TRUE;

    t_dep_desc*  pDD;
    tCC*         pz_ver;
    size_t       ver_len;

    find_version( pz_file, &pz_ver, &ver_len );
    sprintf( pz_b2, "%s/%s", pz_project, pz_file );
    pDD = find_predecessor( pz_b2, pz_ver, ver_len );

    if (pDD == NULL)
        return B_FALSE;

    /*
     *  Check for any prior predecessors.  If none, then
     *  we must set up our work directory.
     */
    if (setup_needed) {
        setup_work_dir( pz_ver );
        setup_needed = B_FALSE;
    }

    mark_targets( pz_file, pDD );
    return B_TRUE;
}


/*
 *  initialize
 *
 *  Find out the names of a number of different directories:
 *  1.  the CVSROOT directory
 *  2.  The official project directory under the root
 *  3.  The user's root directory
 *  4.  the user's project directory (which may be aliased to something
 *      *other* than the official project directory)
 *  5.  a really *large* estimate for the maximum size of a command.
 *      Allocate two contiguous buffers of this size.
 */
    void
initialize( size_t max_file_name )
{
    size_t  max_dir_name = strlen( pz_cvsroot );
    char    z[ 4096 ];

    /*
     *  Separate the $CVSROOT directory from the project directory
     */
    {
        tSCC zRep[]       = "CVS/Root";
        tSCC zNotRooted[] = "cvs dir not under cvs root:\n"
                            "\t%s\n\t%s\n";
        tSCC zBadRoot[]   = "unexpected root format:  %s\n";

        FILE* fp = fopen( zRep, "r" );

        /*
         *  The file must be openable and we have to be able to read
         *  a line from it.  There should be only one line.
         */
        if (fp == (FILE*)NULL) {
            sprintf( z, zFsErr, __LINE__, errno, strerror( errno ),
                     "open-read", zRep );
            fail_exit( z, __LINE__ );
        }

        {
            char* pz = fgets( z, sizeof(z), fp );
            if (pz == (char*)NULL) {
                sprintf( z, zFsErr, __LINE__, errno, strerror( errno ),
                         "reading", zRep );
                fail_exit( z, __LINE__ );
            }

            /*
             *  Close the file and strip trailing white space from the line.
             */
            fclose( fp );
            pz += strlen( pz );
            while ((pz > z) && isspace( pz[-1] ))  pz--;
            *pz = '\0';
        }

        /*
         *  Ensure that our repository directory matches
         */
        {
            size_t  len = strlen( z );

            /*
             *  Check for a remote access preamble to the directory.
             *  We're not interested in that.
             */
            if (z[0] != '/') {
                sprintf( z + len + 1, zBadRoot, z );
                fail_exit( z + len + 1, __LINE__ );
            }

            if (strncmp( pz_cvsroot, z, len ) != 0) {
                sprintf( z+len+1, zNotRooted, z, pz_cvsroot );
                fail_exit( z+len+1, __LINE__ );
            }

            pz_project = pz_cvsroot + len;
            if (*pz_project != '/') {
                sprintf( z+len+1, zNotRooted, z, pz_cvsroot );
                fail_exit( z+len+1, __LINE__ );
            }
        }
    }

    /*
     *  Separate the cvsroot from the project name string
     */
    *(pz_project++) = '\0';

    /*
     *  Now figure out the current user's root and project directories
     */
    pz_myroot = getcwd( NULL, 0 );
    if (pz_myroot == NULL) {
        tSCC zBadDir[] = "fs error %d (%s):  cannot get current dir\n";
        sprintf( z, zBadDir, errno, strerror( errno ));
        fail_exit( z, __LINE__ );
    }

    {
        size_t len = strlen( pz_myroot );
        if (len > max_dir_name)
            max_dir_name = len;
    }

    {
        tSCC zShallow[] = "current directory too shallow:\n"
            "\t%s cannot contain\n\t%s\n";

        char* pz_scan;
        int   dir_ct = 1;

        /*
         *  Count the number of directories in the CVSROOT project directory
         */
        pz_scan = pz_project;
        for (;;) {
            switch (*(pz_scan++)) {
            case '/':
                dir_ct++;
                break;

            case '\0':
                goto dir_ct_done;
            }
        } dir_ct_done:;

        /*
         *  Separate that many directories from the user's root dir
         */
        pz_scan = pz_myroot + strlen( pz_myroot );
        while (dir_ct > 0) {
            if (*(--pz_scan) == '/')
                dir_ct--;
            if (pz_scan == pz_myroot) {
                sprintf( z, zShallow, pz_myroot, pz_project );
                fail_exit( z, __LINE__ );
            }
        }

        *(pz_scan++) = '\0';
        pz_myproj = pz_scan;
    }

    /*
     *  Make really sure we have two buffers big enough to format
     *  whatever commands we want to format.
     */
    max_file_name += max_dir_name + 8 /* more slop */;
    max_file_name *= MAX_FILES_IN_FMT;
    max_file_name += MAX_FORMAT_SIZE;
    max_file_name  = (max_file_name + 0x1000) & ~0x0FFF;
    pz_b1 = (char*)xmalloc( max_file_name * 2 );
    pz_b2 = pz_b1 + max_file_name;
    atexit( cleanup );

    pz_entries = map_data( "CVS/Entries" );
}


/*
 *  map_data
 *
 *  Map in a file and stash all the information needed to unmap it.
 *  Of course, in this program, that is not done  :-}
 */
    tCC*
map_data( tCC* pz_file )
{
    static tCC zCannot[] =
        "fs error %d (%s):  cannot %s %s\n";

    caddr_t res;
    struct stat stbf;
    int fd;

    if (stat( pz_file, &stbf ) != 0)
    {
        sprintf( pz_b2, zCannot, errno, strerror( errno ),
                 "stat", pz_file );
        fail_exit( pz_b2, __LINE__ );
    }

    fd = open( pz_file, O_RDONLY );
    if (fd == -1)
    {
        sprintf( pz_b2, zCannot, errno, strerror( errno ),
                 "open for read", pz_file );
        fail_exit( pz_b2, __LINE__ );
    }

    res = mmap( NULL, stbf.st_size+1, PROT_READ, MAP_SHARED, fd, 0L );
    if (res == MAP_FAILED) {
        sprintf( pz_b2, zCannot, errno, strerror( errno ),
                 "mmap for read/shared", pz_file );
        fail_exit( pz_b2, __LINE__ );
    }

    if (++map_ct > map_allocs) {
        map_allocs += 4;
        p_map_info = (t_map_info*)xrealloc( (void*)p_map_info,
                                            map_allocs * sizeof( *p_map_info ));
    }

    p_map_info[ map_ct-1 ].map_size = stbf.st_size+1;
    p_map_info[ map_ct-1 ].map_dta  = (tCC*)res;
    p_map_info[ map_ct-1 ].map_fd   = fd;

    return (tCC*)res;
}


/*
 *  mark_targets
 *
 *  We do not want to build each target more than once.  Therefore,
 *  increment a reference count for each target that has a predecessor
 *  that is being committed.  While we are at it, we also copy the
 *  predecessor file into our work directory.
 */
    void
mark_targets( tCC*  pz_file, t_dep_desc* pDT )
{
    tSCC zCopyCmd[] =
#ifdef DEBUG
        "set -x\n"
#endif
        "if [ ! -d %1$s/%2$s ] ; then\n"
        "  mkdir -p %1$s/%2$s || echo mkdir FAILED > %4$s\n"
        "fi\ncp  %3$s %1$s/%2$s || echo cp FAILED > %4$s\n";

    tCC**        papz_targ;
    tCC*         pz_targ;
    t_dep_desc*  pTD;

    /*
     *  Copy the file being checked in into the output directory
     */
    sprintf( pz_b1, zCopyCmd, pz_work_dir, pz_project, pz_file, pz_failure );
    run_system( pz_b1, __LINE__ );

    /*
     *  FOR each target file derived from this file being checked in,
     *  DO mark the target->dependencies entry
     */
    papz_targ =   pDT->papz_assoc;
    for (;;) {
        pz_targ = *(papz_targ++);
        if (pz_targ == NULL)
            break;

        pTD = find_dep_entry( target_to_deps, TARGET_TO_DEPS_CT, pz_targ );
        if (pTD == NULL) {
            sprintf( pz_b2, "No dependent entry for target %s\n", pz_targ );
            fail_exit( pz_b2, __LINE__ );
        }

        pTD->referenced++;
    }
}


/*
 *  rebuild_referenced_targets
 *
 *  The main task of the whole program.  For each possibly modified target,
 *  fetch all the predecessor files, cvs get and edit the file and then
 *  emit the instructions for rebuilding the file.  CVS will not commit
 *  the file if it is unchanged, so we do not bother to check.
 */
    void
rebuild_referenced_targets( void )
{
    tSCC zFFun[] =
#ifdef DEBUG
        "set -x\n"
#endif
        "failfile='%s'\n"
        "failfunc() {\n"
        " echo FAILED $@ > ${failfile}\n"
        " exit 1\n}\n";

    tSCC zCommand[] = "cvs -d %s %s %s %s ";
    tSCC zFail[]    = "\\\n  || failfunc exited $? -- %s\n";
    tSCC zCommit[]  =
#ifdef DEBUG
        "set -x\n"
#endif
        "cvs -d %s commit -m'automated commit for %s' %s";

    t_dep_desc*  pTD;
    int          ct;
    size_t       off;  /* sprint offset */
    tCC**        papz;
    char*        pz;
    tCC*         pzc;
    char*        pz_commit_list = pz_b2;
    int          idx;

    if (chdir( pz_work_dir ) != 0) {
        sprintf( pz_b2, zFsErr, __LINE__, errno, strerror( errno ),
                 "chdir", pz_work_dir );
        fail_exit( pz_b2, __LINE__ );
    }

    for (ct = TARGET_TO_DEPS_CT, pTD = target_to_deps;
         ct-- > 0;
         pTD++)  {
        char*  pzSave;

        if (pTD->referenced == 0)
            continue;

        /*
         *  Print the target file name.  We will check it in at the
         *  end of this loop.  However, there may be no rule for it, if
         *  it is a secondary target.
         */
        pz_commit_list += sprintf( pz_commit_list, "%s ", pTD->pz_name );

        /*
         *  Make sure the predecessor files either have been copied from
         *  the commit directory, or have been gotten via "cvs get".
         *  Save and restore the globals "pz_b1" and "pz_b2".
         */
        pzSave = pz_b2;
        pz_b2  = pz_commit_list+1;
        get_predecessor_files( pTD->papz_assoc );
        pz_b2  = pzSave;
    }

    pz = pz_b1 + sprintf( pz_b1, zFFun, pz_failure );
    pz += sprintf( pz, zCommand, pz_cvsroot, "get", pz_ver_tag, pz_b2 );
    pz += sprintf( pz, zFail, "cvs get of targets" );
    pz += sprintf( pz, zCommand, pz_cvsroot, "edit", "", pz_b2 );
    pz += sprintf( pz, zFail, "cvs edit of targets" );
    run_system( pz_b1, __LINE__ );

    for (off = 0, idx = 0, pTD = target_to_deps;
         idx < TARGET_TO_DEPS_CT;
         pTD++, idx++)  {

        /*
         *  ONLY if there is a target rule for this target do we emit build
         *  rules.  There are not target rules for secondary targets
         *  (side effect targets).
         */
        if ((pTD->referenced == 0) || (target_rules[ idx ] == NULL))
            continue;

        /*
         *  FOR each predecessor file,
         *    make it easy to find by the rule script.
         */
        papz = pTD->papz_assoc;
        for (;;) {
            pzc = *(papz++);
            if (pzc == NULL)
                break;
            off += export_file_name( pz_b1 + off, pzc );
        }
    
        /*
         *  We run the rule script in the directory where the
         *  output is to appear.  Extract the directory name from the
         *  target file name.
         */
        pzc = strrchr( pTD->pz_name, '/' );
        if (pzc != NULL) {
            strcpy( pz_b1 + off, "cd " );
            off += 3;
            memcpy( pz_b1 + off, pTD->pz_name, pzc - pTD->pz_name );
            off += (pz - pTD->pz_name);
            pz_b1[off++] = '\n';
        }

        sprintf( pz_b1 + off, "%s\n", target_rules[ idx ]);

        /*
         *  Finally, run each build command.
         */
        run_system( pz_b1, __LINE__ );
    }

    /*
     *  pull a name from the password file to give some attribution in
     *  the commit comment.  We have also carefully preserved the list
     *  of files to commit in pz_b2
     */
    {
        struct passwd * p = getpwuid( getuid() );

        if ((p->pw_gecos != NULL) && (p->pw_gecos[0] != '\0'))
             pz = p->pw_gecos;
        else pz = p->pw_name;

        off = sprintf( pz_b1, zCommit, pz_cvsroot, pz, pz_b2 );
    }

    /*
     *  Commit all the generated files all at once
     */
    sprintf( pz_b1+off, zFail, "commit" );
    run_system( pz_b1, __LINE__ );
}


/*
 *  Run a system command being really verbose if there is a problem
 */
    void
run_system( tCC*  pz_cmd, int line )
{
    tSCC zCmdFail[] = "***  COMMAND FAILED on line %d ***\n";

    fprintf( log_fp, "\n#\n#  RUNNING SCRIPT:\n%s\n\n#  RESULTS:\n", pz_cmd );
    fflush( log_fp );
    fflush( stdout );
    fflush( stderr );

    if (  (system( pz_b1 ) == 0)
       && (access( pz_failure, R_OK ) != 0))  {
        fputs( "\n#\n#  command done\n\n", log_fp );
        return;
    }

    fprintf( log_fp, zCmdFail, line );
    fclose( log_fp );
    fail_exit( "COMMAND FAILURE", line );
}


/*
 *  setup_work_dir
 *
 *  The work directory will be some temporary directory at the root,
 *  followed by the name of the tag being used.  We must serialize
 *  within a tag to ensure we do this in the right order.  For top
 *  of tree, we use a name that we know will never be used as a CVS
 *  tag:  "regen-top-of-tree".
 */
    void
setup_work_dir( tCC* pz_ver )
{
    char*  pz = getenv( "TEMP" );
    size_t tmp_size, sz;
    int fd;

    if (pz != (char*)NULL)
        pz_work_dir = pz;

    if (*pz_ver == '\0')
        pz_ver = "regen-top-of-tree";

    /*
     *  Make sure the base temporary directory exists
     */
    mkdir( pz_work_dir, 0777 );
    tmp_size = sprintf( pz_b1, "%s/%s", pz_work_dir, pz_ver );
    mkdir( pz_b1, 0777 );

    /*
     *  Put our pid at the end of the queue file.  We'll check that
     *  it is top of queue before we start and sleep if someone is
     *  ahead of us.
     */
    strcpy( pz_b1 + tmp_size, "/.queue" );
    if (access( pz_b1, W_OK ) == 0)
         fd = O_APPEND;
    else fd = O_APPEND | O_CREAT;

    fd = open( pz_b1, fd, 02666 );
    if (fd == -1) {
        sprintf( pz_b2, zFsErr, __LINE__, errno, strerror( errno ),
                 "open", pz_b1 );
        fail_exit( pz_b2, __LINE__ );
    }
    sz = sprintf( pz_b2, "%d\n", getpid() );
    write( fd, pz_b2, sz );
    close( fd );
    chmod( pz_b1, 02666 );
    pz_queue = xstrdup( pz_b1 );

    /*
     *  Now create a private work directory where we do everything.
     *  We do not want to worry about other processes copying files around.
     */
    sz = sprintf( pz_b1 + tmp_size, "/work%d", getpid() );
    if (mkdir( pz_b1, 0777 ) != 0) {
        sprintf( pz_b2, zFsErr, __LINE__, errno, strerror( errno ),
                 "mkdir", pz_b1 );
        fail_exit( pz_b2, __LINE__ );
    }

    /*
     *  Publish the name of the work directory
     */
    tmp_size += sz + 1;
    pz_work_dir = (char*)xmalloc( tmp_size );
    strcpy( pz_work_dir, pz_b1 );
    {
        tSCC zFailFmt[] = "%s/FAILED";
        pz_failure = (char*)xmalloc( tmp_size + sizeof(zFailFmt) );
        sprintf( pz_failure, zFailFmt, pz_b1 );
    }
    strcpy( pz_b1+tmp_size-1, "/LOG" );
    log_fp = fopen( pz_b1, "w" );
}


/* regen.c ends here */
